using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Xml;

namespace RayCasting2D {

	public class Scene {

		private List<Tile> tiles = new List<Tile>();
		private int horizontalTileCount = 32;
		private int verticalTileCount = 32;
		private int tileSize;

		public Scene(int horizontalTileCount, int verticalTileCount, int tileSize) {
			this.horizontalTileCount = horizontalTileCount;
			this.verticalTileCount = verticalTileCount;
			this.tileSize = tileSize;

			while (tiles.Count < horizontalTileCount * verticalTileCount) {
				tiles.Add(null);
			}
		}

		public void SetTile(int x, int y, Tile tile) {
			tiles[y * horizontalTileCount + x] = tile;
		}

		public Tile GetTile(int x, int y) {
			return tiles[y * horizontalTileCount + x];
		}
		
		// Find out if the given pixel is traversable.
		// X and Y are the scene pixel coordinates
		public bool IsPointTraversable(int x, int y) {
			// Get the tile coordinates from the pixel coord.
			int tileX = (int)Math.Floor((float)x / (float)tileSize);
			int tileY = (int)Math.Floor((float)y / (float)tileSize);

			// If the point is out of bound, we assume it's traversable
			// Although it might depend on the game
			if (tileX < 0) return true;
			if (tileX >= horizontalTileCount) return true;
			if (tileY < 0) return true;
			if (tileY >= verticalTileCount) return true;

			Tile tile = GetTile(tileX, tileY);

			// If the tile is blank the point is traversable
			if (tile == null) return true;

			// Get the coordinates of the point within the tile
			int localPointX = x % tileSize;
			int localPointY = y % tileSize;

			// Return "true" if the pixel is not solid
			return !tile.SolidityMap[localPointX, localPointY];
		}
		
		// Class returned by the RayCast() method
		public class RayCastingResult {
			// Does the ray collide with the environment?
			private bool doCollide;
			// And if so, at which position?
			private Vector position;
			

			public bool DoCollide {
				get { return doCollide; }
				set { doCollide = value; }
			}

			public Vector Position {
				get { return position; }
				set { position = value; }
			}
		}

		public RayCastingResult RayCast(Vector position, Vector direction, float rayLength) {
			RayCastingResult result = new RayCastingResult();
			result.DoCollide = false;

			// Exit the function now if the ray length is 0
			if (rayLength == 0) {
				result.DoCollide = IsPointTraversable((int)position.X, (int)position.Y);
				result.Position = position;

				return result;
			}

			// Get the list of points from the Bresenham algorithm
			List<Point> rayLine = BresenhamLine(position, position + (direction.GetNormalized() * rayLength));

			if (rayLine.Count > 0) {
				int rayPointIndex = 0;

				if (rayLine[0] != position) rayPointIndex = rayLine.Count - 1;

				// Loop through all the points starting from "position"
				while (true) {
					Point rayPoint = rayLine[rayPointIndex];
					if (!IsPointTraversable(rayPoint.X, rayPoint.Y)) {
						result.Position = Vector.FromPoint(rayPoint);
						result.DoCollide = true;
						break;
					}
					if (rayLine[0] != position) {
						rayPointIndex--;
						if (rayPointIndex < 0) break;
					} else {
						rayPointIndex++;
						if (rayPointIndex >= rayLine.Count) break;
					}
				}
			}

			return result;
		}

		// Swap the values of A and B
		private void Swap<T>(ref T a, ref T b) {
			T c = a;
			a = b;
			b = c;
		}

		// Returns the list of points from p0 to p1 
		private List<Point> BresenhamLine(Point p0, Point p1) {
			return BresenhamLine(p0.X, p0.Y, p1.X, p1.Y);
		}

		// Returns the list of points from (x0, y0) to (x1, y1)
		private List<Point> BresenhamLine(int x0, int y0, int x1, int y1) {
			// Optimization: it would be preferable to calculate in
			// advance the size of "result" and to use a fixed-size array
			// instead of a list.
			List<Point> result = new List<Point>();

			bool steep = Math.Abs(y1 - y0) > Math.Abs(x1 - x0);
			if (steep) {
				Swap(ref x0, ref y0);
				Swap(ref x1, ref y1);
			}
			if (x0 > x1) {
				Swap(ref x0, ref x1);
				Swap(ref y0, ref y1);
			}

			int deltax = x1 - x0;
			int deltay = Math.Abs(y1 - y0);
			int error = 0;
			int ystep;
			int y = y0;
			if (y0 < y1) ystep = 1; else ystep = -1;
			for (int x = x0; x <= x1; x++) {
				if (steep) result.Add(new Point(y, x)); else result.Add(new Point(x, y));
				error += deltay;
				if (2 * error >= deltax) {
					y += ystep;
					error -= deltax;
				}
			}

			return result;
		}

		public void Draw(Graphics g) {
			for (int tileX = 0; tileX < horizontalTileCount; tileX++) {
				for (int tileY = 0; tileY < verticalTileCount; tileY++) {
					Tile tile = tiles[tileY * horizontalTileCount + tileX];
					if (tile == null) continue;
					Rectangle sourceRect = new Rectangle(0, 0, tile.Bitmap.Width, tile.Bitmap.Height);

					if (tile.HorizontalFlip) {
						sourceRect.X = sourceRect.X + sourceRect.Width;
						sourceRect.Width = -sourceRect.Width;
					}
					if (tile.VerticalFlip) {
						sourceRect.Y = sourceRect.Y + sourceRect.Height;
						sourceRect.Height = -sourceRect.Height;
					}

					Rectangle destRect = new Rectangle(tileX * tileSize, tileY * tileSize, tileSize, tileSize);

					g.DrawImage(tile.Bitmap, destRect, sourceRect, GraphicsUnit.Pixel);
				}
			}
		}

	}

}
